@@ -30,6 +30,8 @@
 
 #ifdef HAVE_OVIRT
 #include <govirt/govirt.h>
+#include "ovirt-foreign-menu.h"
+#include "virt-viewer-vm-connection.h"
 #endif
 
 #ifdef HAVE_SPICE_GTK
@@ -44,6 +46,7 @@
 #include "virt-viewer-file.h"
 #include "virt-viewer-session.h"
 #include "remote-viewer.h"
+#include "remote-viewer-connect.h"
 
 #ifndef G_VALUE_INIT /* see bug https://bugzilla.gnome.org/show_bug.cgi?id=654793 */
 #define G_VALUE_INIT  { 0, { { 0 } } }
@@ -54,8 +57,9 @@ struct _RemoteViewerPrivate {
     SpiceCtrlController *controller;
     SpiceCtrlForeignMenu *ctrl_foreign_menu;
 #endif
-    GtkWidget *controller_menu;
-    GtkWidget *foreign_menu;
+#ifdef HAVE_OVIRT
+    OvirtForeignMenu *ovirt_foreign_menu;
+#endif
     gboolean open_recent_dialog;
 };
 
@@ -72,19 +76,29 @@ enum {
     PROP_OPEN_RECENT_DIALOG
 };
 
-static gboolean remote_viewer_start(VirtViewerApp *self);
+#ifdef HAVE_OVIRT
+static OvirtVm * choose_vm(GtkWindow *main_window,
+                           char **vm_name,
+                           OvirtCollection *vms,
+                           GError **error);
+#endif
+
+static gboolean remote_viewer_start(VirtViewerApp *self, GError **error);
 #ifdef HAVE_SPICE_GTK
 static gboolean remote_viewer_activate(VirtViewerApp *self, GError **error);
 static void remote_viewer_window_added(VirtViewerApp *self, VirtViewerWindow *win);
 static void spice_foreign_menu_updated(RemoteViewer *self);
-static gint connect_dialog(gchar **uri);
+#endif
 
 static void
 remote_viewer_dispose (GObject *object)
 {
+#if defined(HAVE_SPICE_GTK) || defined(HAVE_OVIRT)
     RemoteViewer *self = REMOTE_VIEWER(object);
     RemoteViewerPrivate *priv = self->priv;
+#endif
 
+#ifdef HAVE_SPICE_GTK
     if (priv->controller) {
         g_object_unref(priv->controller);
         priv->controller = NULL;
@@ -94,10 +108,17 @@ remote_viewer_dispose (GObject *object)
         g_object_unref(priv->ctrl_foreign_menu);
         priv->ctrl_foreign_menu = NULL;
     }
+#endif
+
+#ifdef HAVE_OVIRT
+    if (priv->ovirt_foreign_menu) {
+        g_object_unref(priv->ovirt_foreign_menu);
+        priv->ovirt_foreign_menu = NULL;
+    }
+#endif
 
     G_OBJECT_CLASS(remote_viewer_parent_class)->dispose (object);
 }
-#endif
 
 static void
 remote_viewer_get_property (GObject *object, guint property_id,
@@ -156,7 +177,7 @@ remote_viewer_deactivated(VirtViewerApp *app, gboolean connect_error)
     RemoteViewerPrivate *priv = self->priv;
 
     if (connect_error && priv->open_recent_dialog) {
-        if (virt_viewer_app_start(app)) {
+        if (virt_viewer_app_start(app, NULL)) {
             return;
         }
     }
@@ -177,13 +198,11 @@ remote_viewer_class_init (RemoteViewerClass *klass)
 
     app_class->start = remote_viewer_start;
     app_class->deactivated = remote_viewer_deactivated;
-#ifdef HAVE_SPICE_GTK
     object_class->dispose = remote_viewer_dispose;
+#ifdef HAVE_SPICE_GTK
     app_class->activate = remote_viewer_activate;
     app_class->window_added = remote_viewer_window_added;
-#endif
 
-#ifdef HAVE_SPICE_GTK
     g_object_class_install_property(object_class,
                                     PROP_CONTROLLER,
                                     g_param_spec_object("controller",
@@ -396,8 +415,7 @@ spice_menu_update(RemoteViewer *self, VirtViewerWindow *win)
 }
 
 static void
-spice_menu_update_each(gpointer key G_GNUC_UNUSED,
-                       gpointer value,
+spice_menu_update_each(gpointer value,
                        gpointer user_data)
 {
     spice_menu_update(REMOTE_VIEWER(user_data), VIRT_VIEWER_WINDOW(value));
@@ -406,15 +424,15 @@ spice_menu_update_each(gpointer key G_GNUC_UNUSED,
 static void
 spice_ctrl_menu_updated(RemoteViewer *self)
 {
-    GHashTable *windows = virt_viewer_app_get_windows(VIRT_VIEWER_APP(self));
+    GList *windows = virt_viewer_app_get_windows(VIRT_VIEWER_APP(self));
 
     g_debug("Spice controller menu updated");
 
-    g_hash_table_foreach(windows, spice_menu_update_each, self);
+    g_list_foreach(windows, spice_menu_update_each, self);
 }
 
 static void
-foreign_menu_update(RemoteViewer *self, VirtViewerWindow *win)
+spice_foreign_menu_update(RemoteViewer *self, VirtViewerWindow *win)
 {
     GtkWidget *menuitem = g_object_get_data(G_OBJECT(win), "foreign-menu");
     SpiceCtrlMenu *menu;
@@ -445,21 +463,20 @@ foreign_menu_update(RemoteViewer *self, VirtViewerWindow *win)
 }
 
 static void
-foreign_menu_update_each(gpointer key G_GNUC_UNUSED,
-                         gpointer value,
-                         gpointer user_data)
+spice_foreign_menu_update_each(gpointer value,
+                               gpointer user_data)
 {
-    foreign_menu_update(REMOTE_VIEWER(user_data), VIRT_VIEWER_WINDOW(value));
+    spice_foreign_menu_update(REMOTE_VIEWER(user_data), VIRT_VIEWER_WINDOW(value));
 }
 
 static void
 spice_foreign_menu_updated(RemoteViewer *self)
 {
-    GHashTable *windows = virt_viewer_app_get_windows(VIRT_VIEWER_APP(self));
+    GList *windows = virt_viewer_app_get_windows(VIRT_VIEWER_APP(self));
 
     g_debug("Spice foreign menu updated");
 
-    g_hash_table_foreach(windows, foreign_menu_update_each, self);
+    g_list_foreach(windows, spice_foreign_menu_update_each, self);
 }
 
 static SpiceSession *
@@ -616,17 +633,17 @@ remote_viewer_activate(VirtViewerApp *app, GError **error)
 }
 
 static void
-remote_viewer_window_added(VirtViewerApp *app G_GNUC_UNUSED,
+remote_viewer_window_added(VirtViewerApp *app,
                            VirtViewerWindow *win)
 {
     spice_menu_update(REMOTE_VIEWER(app), win);
-    foreign_menu_update(REMOTE_VIEWER(app), win);
+    spice_foreign_menu_update(REMOTE_VIEWER(app), win);
 }
 #endif
 
 #ifdef HAVE_OVIRT
 static gboolean
-parse_ovirt_uri(const gchar *uri_str, char **rest_uri, char **name)
+parse_ovirt_uri(const gchar *uri_str, char **rest_uri, char **name, char **username)
 {
     char *vm_name = NULL;
     char *rel_path;
@@ -639,15 +656,24 @@ parse_ovirt_uri(const gchar *uri_str, char **rest_uri, char **name)
     g_return_val_if_fail(name != NULL, FALSE);
 
     uri = xmlParseURI(uri_str);
-    if (uri == NULL)
-        return FALSE;
+    g_return_val_if_fail(uri != NULL, FALSE);
 
     if (g_strcmp0(uri->scheme, "ovirt") != 0) {
         xmlFreeURI(uri);
         return FALSE;
     }
 
+    if (username && uri->user)
+        *username = g_strdup(uri->user);
+
     if (uri->path == NULL) {
+        *name = NULL;
+        *rest_uri = g_strdup(uri->server);
+        xmlFreeURI(uri);
+        return TRUE;
+    }
+
+    if (*uri->path != '/') {
         xmlFreeURI(uri);
         return FALSE;
     }
@@ -658,7 +684,6 @@ parse_ovirt_uri(const gchar *uri_str, char **rest_uri, char **name)
     element_count = g_strv_length(path_elements);
     if (element_count == 0) {
         g_strfreev(path_elements);
-        xmlFreeURI(uri);
         return FALSE;
     }
     vm_name = path_elements[element_count-1];
@@ -666,8 +691,7 @@ parse_ovirt_uri(const gchar *uri_str, char **rest_uri, char **name)
 
     /* build final URI */
     rel_path = g_strjoinv("/", path_elements);
-    /* FIXME: how to decide between http and https? */
-    *rest_uri = g_strdup_printf("https://%s/%s/api/", uri->server, rel_path);
+    *rest_uri = g_strdup_printf("%s%s", uri->server, rel_path);
     *name = vm_name;
     g_free(rel_path);
     g_strfreev(path_elements);
@@ -683,31 +707,122 @@ static gboolean
 authenticate_cb(RestProxy *proxy, G_GNUC_UNUSED RestProxyAuth *auth,
                 G_GNUC_UNUSED gboolean retrying, gpointer user_data)
 {
-    gchar *username;
-    gchar *password;
+    gchar *username = NULL;
+    gchar *password = NULL;
     VirtViewerWindow *window;
+    gboolean success = FALSE;
+
+    g_object_get(proxy,
+                 "username", &username,
+                 NULL);
+
+    if (username == NULL || *username == '\0')
+        username = g_strdup(g_get_user_name());
 
     window = virt_viewer_app_get_main_window(VIRT_VIEWER_APP(user_data));
-    int ret = virt_viewer_auth_collect_credentials(virt_viewer_window_get_window(window),
+    success = virt_viewer_auth_collect_credentials(virt_viewer_window_get_window(window),
                                                    "oVirt",
                                                    NULL,
                                                    &username, &password);
-    if (ret < 0) {
-        return FALSE;
-    } else {
+    if (success) {
         g_object_set(G_OBJECT(proxy),
                      "username", username,
                      "password", password,
                      NULL);
-        g_free(username);
-        g_free(password);
-        return TRUE;
+#ifdef HAVE_OVIRT_CANCEL
+    } else {
+        rest_proxy_auth_cancel(auth);
+#endif
     }
+
+    g_free(username);
+    g_free(password);
+    return success;
+}
+
+static void
+ovirt_foreign_menu_update(RemoteViewer *app, VirtViewerWindow *win)
+{
+    GtkWidget *menu = g_object_get_data(G_OBJECT(win), "foreign-menu");
+    GtkWidget *submenu;
+    GtkMenuShell *shell = GTK_MENU_SHELL(gtk_builder_get_object(virt_viewer_window_get_builder(win), "top-menu"));
+
+    if (app->priv->ovirt_foreign_menu == NULL) {
+        /* nothing to do */
+        return;
+    }
+    if (menu == NULL) {
+        menu = gtk_menu_item_new_with_label(_("_Change CD"));
+        gtk_menu_item_set_use_underline(GTK_MENU_ITEM(menu), TRUE);
+        gtk_menu_shell_append(shell, menu);
+        g_object_set_data_full(G_OBJECT(win), "foreign-menu",
+                               g_object_ref(menu),
+                               (GDestroyNotify)gtk_widget_destroy);
+    }
+
+    submenu = ovirt_foreign_menu_get_gtk_menu(app->priv->ovirt_foreign_menu);
+    if (submenu != NULL) {
+        gtk_menu_item_set_submenu(GTK_MENU_ITEM(menu), submenu);
+    } else {
+        /* No items to show, no point in showing the menu */
+        g_object_set_data(G_OBJECT(win), "foreign-menu", NULL);
+    }
+
+    gtk_widget_show_all(menu);
+}
+
+static void
+ovirt_foreign_menu_update_each(gpointer value,
+                               gpointer user_data)
+{
+    ovirt_foreign_menu_update(REMOTE_VIEWER(user_data),
+                              VIRT_VIEWER_WINDOW(value));
+}
+
+static void
+ovirt_foreign_menu_updated(RemoteViewer *self)
+{
+    GList *windows = virt_viewer_app_get_windows(VIRT_VIEWER_APP(self));
+
+    g_debug("Spice foreign menu updated");
+
+    g_list_foreach(windows, ovirt_foreign_menu_update_each, self);
+}
+
+static void
+ovirt_foreign_menu_changed(OvirtForeignMenu *foreign_menu G_GNUC_UNUSED,
+                           GParamSpec *pspec G_GNUC_UNUSED,
+                           VirtViewerApp *app)
+{
+    ovirt_foreign_menu_updated(REMOTE_VIEWER(app));
+}
+
+
+static void
+virt_viewer_app_set_ovirt_foreign_menu(VirtViewerApp *app,
+                                       OvirtForeignMenu *foreign_menu)
+{
+    RemoteViewer *self;
+    g_return_if_fail(REMOTE_VIEWER_IS(app));
+    g_return_if_fail(OVIRT_IS_FOREIGN_MENU(foreign_menu));
+
+    self = REMOTE_VIEWER(app);
+    if (self->priv->ovirt_foreign_menu != NULL) {
+        g_object_unref(G_OBJECT(self->priv->ovirt_foreign_menu));
+    }
+    self->priv->ovirt_foreign_menu = foreign_menu;
+    g_signal_connect(G_OBJECT(foreign_menu), "notify::file",
+                     (GCallback)ovirt_foreign_menu_changed, app);
+    g_signal_connect(G_OBJECT(foreign_menu), "notify::files",
+                     (GCallback)ovirt_foreign_menu_changed, app);
+    g_signal_connect(G_OBJECT(app), "window-added",
+                     (GCallback)ovirt_foreign_menu_update, NULL);
+    ovirt_foreign_menu_start(foreign_menu);
 }
 
 
 static gboolean
-create_ovirt_session(VirtViewerApp *app, const char *uri)
+create_ovirt_session(VirtViewerApp *app, const char *uri, GError **err)
 {
     OvirtProxy *proxy = NULL;
     OvirtApi *api = NULL;
@@ -718,9 +833,11 @@ create_ovirt_session(VirtViewerApp *app, const char *uri)
     GError *error = NULL;
     char *rest_uri = NULL;
     char *vm_name = NULL;
+    char *username = NULL;
     gboolean success = FALSE;
     guint port;
     guint secure_port;
+    char *proxy_url = NULL;
     OvirtVmDisplayType type;
     const char *session_type;
 
@@ -729,14 +846,20 @@ create_ovirt_session(VirtViewerApp *app, const char *uri)
     gchar *ghost = NULL;
     gchar *ticket = NULL;
     gchar *host_subject = NULL;
+    gchar *guid = NULL;
 
     g_return_val_if_fail(VIRT_VIEWER_IS_APP(app), FALSE);
 
-    if (!parse_ovirt_uri(uri, &rest_uri, &vm_name))
+    if (!parse_ovirt_uri(uri, &rest_uri, &vm_name, &username)) {
+        g_set_error_literal(&error, VIRT_VIEWER_ERROR, VIRT_VIEWER_ERROR_FAILED,
+                            _("failed to parse ovirt uri"));
         goto error;
+    }
+
     proxy = ovirt_proxy_new(rest_uri);
-    if (proxy == NULL)
-        goto error;
+    g_object_set(proxy,
+                 "username", username,
+                 NULL);
     ovirt_set_proxy_options(proxy);
     g_signal_connect(G_OBJECT(proxy), "authenticate",
                      G_CALLBACK(authenticate_cb), app);
@@ -744,6 +867,14 @@ create_ovirt_session(VirtViewerApp *app, const char *uri)
     api = ovirt_proxy_fetch_api(proxy, &error);
     if (error != NULL) {
         g_debug("failed to get oVirt 'api' collection: %s", error->message);
+#ifdef HAVE_OVIRT_CANCEL
+        if (g_error_matches(error, OVIRT_REST_CALL_ERROR, OVIRT_REST_CALL_ERROR_CANCELLED)) {
+            g_clear_error(&error);
+            g_set_error_literal(&error,
+                                VIRT_VIEWER_ERROR, VIRT_VIEWER_ERROR_CANCELLED,
+                                _("Authentication was cancelled"));
+        }
+#endif
         goto error;
     }
     vms = ovirt_api_get_vms(api);
@@ -752,24 +883,42 @@ create_ovirt_session(VirtViewerApp *app, const char *uri)
         g_debug("failed to lookup %s: %s", vm_name, error->message);
         goto error;
     }
-    vm = OVIRT_VM(ovirt_collection_lookup_resource(vms, vm_name));
-    g_return_val_if_fail(vm != NULL, FALSE);
+    if (vm_name == NULL ||
+        (vm = OVIRT_VM(ovirt_collection_lookup_resource(vms, vm_name))) == NULL) {
+        VirtViewerWindow *main_window = virt_viewer_app_get_main_window(app);
+        vm = choose_vm(virt_viewer_window_get_window(main_window),
+                       &vm_name,
+                       vms,
+                       &error);
+        if (vm == NULL) {
+            goto error;
+        }
+    }
     g_object_get(G_OBJECT(vm), "state", &state, NULL);
     if (state != OVIRT_VM_STATE_UP) {
-        g_debug("oVirt VM %s is not running", vm_name);
+        g_set_error(&error, VIRT_VIEWER_ERROR, VIRT_VIEWER_ERROR_FAILED,
+                    _("oVirt VM %s is not running"), vm_name);
+        g_debug("%s", error->message);
         goto error;
     }
+    g_object_set(app, "guest-name", vm_name, NULL);
 
     if (!ovirt_vm_get_ticket(vm, proxy, &error)) {
         g_debug("failed to get ticket for %s: %s", vm_name, error->message);
         goto error;
     }
 
-    g_object_get(G_OBJECT(vm), "display", &display, NULL);
+    g_object_get(G_OBJECT(vm), "display", &display, "guid", &guid, NULL);
     if (display == NULL) {
+        g_set_error(&error, VIRT_VIEWER_ERROR, VIRT_VIEWER_ERROR_FAILED,
+                    _("oVirt VM %s has no display"), vm_name);
         goto error;
     }
 
+    if (guid != NULL) {
+        g_object_set(app, "uuid", guid, NULL);
+    }
+
     g_object_get(G_OBJECT(display),
                  "type", &type,
                  "address", &ghost,
@@ -777,23 +926,43 @@ create_ovirt_session(VirtViewerApp *app, const char *uri)
                  "secure-port", &secure_port,
                  "ticket", &ticket,
                  "host-subject", &host_subject,
+                 "proxy-url", &proxy_url,
                  NULL);
-    gport = g_strdup_printf("%d", port);
-    gtlsport = g_strdup_printf("%d", secure_port);
+    if (port != 0) {
+        gport = g_strdup_printf("%d", port);
+    }
+    if (secure_port != 0) {
+        gtlsport = g_strdup_printf("%d", secure_port);
+    }
+
+    if (ghost == NULL) {
+        g_set_error(&error, VIRT_VIEWER_ERROR, VIRT_VIEWER_ERROR_FAILED,
+                    _("oVirt VM %s has no host information"), vm_name);
+        g_debug("%s", error->message);
+        goto error;
+    }
 
     if (type == OVIRT_VM_DISPLAY_SPICE) {
         session_type = "spice";
     } else if (type == OVIRT_VM_DISPLAY_VNC) {
         session_type = "vnc";
     } else {
-        g_debug("Unknown display type: %d", type);
+        g_set_error(&error, VIRT_VIEWER_ERROR, VIRT_VIEWER_ERROR_FAILED,
+                    _("oVirt VM %s has unknown display type: %d"), vm_name, type);
+        g_debug("%s", error->message);
         goto error;
     }
 
+    {
+        OvirtForeignMenu *ovirt_menu = ovirt_foreign_menu_new(proxy);
+        g_object_set(G_OBJECT(ovirt_menu), "api", api, "vm", vm, NULL);
+        virt_viewer_app_set_ovirt_foreign_menu(app, ovirt_menu);
+    }
+
     virt_viewer_app_set_connect_info(app, NULL, ghost, gport, gtlsport,
                                      session_type, NULL, NULL, 0, NULL);
 
-    if (virt_viewer_app_create_session(app, session_type) < 0)
+    if (!virt_viewer_app_create_session(app, session_type, &error))
         goto error;
 
 #ifdef HAVE_SPICE_GTK
@@ -805,6 +974,7 @@ create_ovirt_session(VirtViewerApp *app, const char *uri)
         g_object_set(G_OBJECT(session),
                      "password", ticket,
                      "cert-subject", host_subject,
+                     "proxy", proxy_url,
                      NULL);
         g_object_get(G_OBJECT(proxy), "ca-cert", &ca_cert, NULL);
         if (ca_cert != NULL) {
@@ -819,6 +989,7 @@ create_ovirt_session(VirtViewerApp *app, const char *uri)
     success = TRUE;
 
 error:
+    g_free(username);
     g_free(rest_uri);
     g_free(vm_name);
     g_free(ticket);
@@ -826,9 +997,11 @@ error:
     g_free(gtlsport);
     g_free(ghost);
     g_free(host_subject);
+    g_free(guid);
+    g_free(proxy_url);
 
     if (error != NULL)
-        g_error_free(error);
+        g_propagate_error(err, error);
     if (display != NULL)
         g_object_unref(display);
     if (vm != NULL)
@@ -843,163 +1016,50 @@ error:
 
 #endif
 
-static void entry_icon_release_cb(GtkEntry* entry, gpointer data G_GNUC_UNUSED)
+#ifdef HAVE_OVIRT
+static OvirtVm *
+choose_vm(GtkWindow *main_window,
+          char **vm_name,
+          OvirtCollection *vms_collection,
+          GError **error)
 {
-    gtk_entry_set_text(entry, "");
-    gtk_widget_grab_focus(GTK_WIDGET(entry));
-}
+    GtkListStore *model;
+    GtkTreeIter iter;
+    GHashTable *vms;
+    GHashTableIter vms_iter;
+    OvirtVmState state;
+    OvirtVm *vm;
 
-static void entry_changed_cb(GtkEditable* entry, gpointer data G_GNUC_UNUSED)
-{
-    gboolean rtl = (gtk_widget_get_direction(GTK_WIDGET(entry)) == GTK_TEXT_DIR_RTL);
-    gboolean active = gtk_entry_get_text_length(GTK_ENTRY(entry)) > 0;
+    g_return_val_if_fail(vm_name != NULL, NULL);
+    free(*vm_name);
 
-    g_object_set(entry,
-                 "secondary-icon-name", active ? (rtl ? "edit-clear-rtl-symbolic" : "edit-clear-symbolic") : NULL,
-                 "secondary-icon-activatable", active,
-                 "secondary-icon-sensitive", active,
-                 NULL);
-}
+    model = gtk_list_store_new(1, G_TYPE_STRING);
 
-static void
-recent_selection_changed_dialog_cb(GtkRecentChooser *chooser, gpointer data)
-{
-    GtkRecentInfo *info;
-    GtkWidget *entry = data;
-    const gchar *uri;
-
-    info = gtk_recent_chooser_get_current_item(chooser);
-    if (info == NULL)
-        return;
-
-    uri = gtk_recent_info_get_uri(info);
-    g_return_if_fail(uri != NULL);
-
-    gtk_entry_set_text(GTK_ENTRY(entry), uri);
-
-    gtk_recent_info_unref(info);
-}
-
-static void
-recent_item_activated_dialog_cb(GtkRecentChooser *chooser G_GNUC_UNUSED, gpointer data)
-{
-   gtk_dialog_response(GTK_DIALOG (data), GTK_RESPONSE_ACCEPT);
-}
-
-static void make_label_light(GtkLabel* label)
-{
-    PangoAttrList* attributes = pango_attr_list_new();
-#if GTK_CHECK_VERSION(3, 0, 0)
-    gtk_style_context_add_class(gtk_widget_get_style_context(GTK_WIDGET(label)), "dim-label");
-#else
-    GtkStyle* style = gtk_widget_get_style(GTK_WIDGET(label));
-    GdkColor* c = &style->text[GTK_STATE_INSENSITIVE];
-    pango_attr_list_insert(attributes, pango_attr_foreground_new(c->red, c->green, c->blue));
-#endif
-    pango_attr_list_insert(attributes, pango_attr_scale_new(0.9));
-    gtk_label_set_attributes(label, attributes);
-    pango_attr_list_unref(attributes);
-}
-
-static void make_label_bold(GtkLabel* label)
-{
-    PangoAttrList* attributes = pango_attr_list_new();
-    pango_attr_list_insert(attributes, pango_attr_weight_new(PANGO_WEIGHT_BOLD));
-    gtk_label_set_attributes(label, attributes);
-    pango_attr_list_unref(attributes);
-
-}
-
-static gint
-connect_dialog(gchar **uri)
-{
-    GtkWidget *dialog, *area, *box, *label, *entry, *recent;
-#if !GTK_CHECK_VERSION(3, 0, 0)
-    GtkWidget *alignment;
-#endif
-    GtkRecentFilter *rfilter;
-    gint retval;
-
-    /* Create the widgets */
-    dialog = gtk_dialog_new_with_buttons(_("Connection details"),
-                                         NULL,
-                                         GTK_DIALOG_DESTROY_WITH_PARENT,
-                                         GTK_STOCK_CANCEL,
-                                         GTK_RESPONSE_REJECT,
-                                         GTK_STOCK_CONNECT,
-                                         GTK_RESPONSE_ACCEPT,
-                                         NULL);
-    gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT);
-    gtk_container_set_border_width(GTK_CONTAINER(dialog), 5);
-    area = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
-    box = gtk_vbox_new(FALSE, 6);
-    gtk_container_set_border_width(GTK_CONTAINER(box), 5);
-    gtk_box_pack_start(GTK_BOX(area), box, TRUE, TRUE, 0);
-
-    label = gtk_label_new_with_mnemonic(_("_Connection Address"));
-    gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
-    gtk_box_pack_start(GTK_BOX(box), label, TRUE, TRUE, 0);
-    entry = GTK_WIDGET(gtk_entry_new());
-    gtk_entry_set_activates_default(GTK_ENTRY(entry), TRUE);
-    g_object_set(entry, "width-request", 200, NULL);
-    g_signal_connect(entry, "changed", G_CALLBACK(entry_changed_cb), entry);
-    g_signal_connect(entry, "icon-release", G_CALLBACK(entry_icon_release_cb), entry);
-    gtk_box_pack_start(GTK_BOX(box), entry, TRUE, TRUE, 0);
-    gtk_label_set_mnemonic_widget(GTK_LABEL(label), entry);
-    make_label_bold(GTK_LABEL(label));
-
-    label = gtk_label_new(_("For example, spice://foo.example.org:5900"));
-    gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
-    make_label_light(GTK_LABEL(label));
-#if GTK_CHECK_VERSION(3, 0, 0)
-    gtk_box_pack_start(GTK_BOX(box), label, TRUE, TRUE, 0);
-    gtk_widget_set_margin_bottom(label, 12);
-#else
-    alignment = gtk_alignment_new(0, 0, 1, 1);
-    gtk_alignment_set_padding(GTK_ALIGNMENT(alignment), 0, 12, 0, 0);
-    gtk_container_add(GTK_CONTAINER(alignment), label);
-    gtk_box_pack_start(GTK_BOX(box), alignment, TRUE, TRUE, 0);
-#endif
-
-    label = gtk_label_new_with_mnemonic(_("_Recent Connections"));
-    make_label_bold(GTK_LABEL(label));
-    gtk_box_pack_start(GTK_BOX(box), label, TRUE, TRUE, 0);
-    gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
-
-    recent = GTK_WIDGET(gtk_recent_chooser_widget_new());
-    gtk_recent_chooser_set_show_icons(GTK_RECENT_CHOOSER(recent), FALSE);
-    gtk_recent_chooser_set_sort_type(GTK_RECENT_CHOOSER(recent), GTK_RECENT_SORT_MRU);
-    gtk_box_pack_start(GTK_BOX(box), recent, TRUE, TRUE, 0);
-    gtk_label_set_mnemonic_widget(GTK_LABEL(label), recent);
-
-    rfilter = gtk_recent_filter_new();
-    gtk_recent_filter_add_mime_type(rfilter, "application/x-spice");
-    gtk_recent_filter_add_mime_type(rfilter, "application/x-vnc");
-    gtk_recent_filter_add_mime_type(rfilter, "application/x-virt-viewer");
-    gtk_recent_chooser_set_filter(GTK_RECENT_CHOOSER(recent), rfilter);
-    gtk_recent_chooser_set_local_only(GTK_RECENT_CHOOSER(recent), FALSE);
-    g_signal_connect(recent, "selection-changed",
-                     G_CALLBACK(recent_selection_changed_dialog_cb), entry);
-    g_signal_connect(recent, "item-activated",
-                     G_CALLBACK(recent_item_activated_dialog_cb), dialog);
-
-    /* show and wait for response */
-    gtk_widget_show_all(dialog);
-    if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT) {
-        *uri = g_strdup(gtk_entry_get_text(GTK_ENTRY(entry)));
-        g_strstrip(*uri);
-        retval = 0;
-    } else {
-        *uri = NULL;
-        retval = -1;
+    vms = ovirt_collection_get_resources(vms_collection);
+    g_hash_table_iter_init(&vms_iter, vms);
+    while (g_hash_table_iter_next(&vms_iter, (gpointer *) vm_name, (gpointer *) &vm)) {
+        g_object_get(G_OBJECT(vm), "state", &state, NULL);
+        if (state == OVIRT_VM_STATE_UP) {
+            gtk_list_store_append(model, &iter);
+            gtk_list_store_set(model, &iter, 0, *vm_name, -1);
+       }
     }
-    gtk_widget_destroy(dialog);
 
-    return retval;
+    *vm_name = virt_viewer_vm_connection_choose_name_dialog(main_window,
+                                                            GTK_TREE_MODEL(model),
+                                                            error);
+    g_object_unref(model);
+    if (*vm_name == NULL)
+        return NULL;
+
+    vm = OVIRT_VM(ovirt_collection_lookup_resource(vms_collection, *vm_name));
+
+    return vm;
 }
+#endif
 
 static gboolean
-remote_viewer_start(VirtViewerApp *app)
+remote_viewer_start(VirtViewerApp *app, GError **err)
 {
     g_return_val_if_fail(REMOTE_VIEWER_IS(app), FALSE);
 
@@ -1016,8 +1076,8 @@ remote_viewer_start(VirtViewerApp *app)
     g_signal_connect(app, "notify", G_CALLBACK(app_notified), self);
 
     if (priv->controller) {
-        if (virt_viewer_app_create_session(app, "spice") < 0) {
-            virt_viewer_app_simple_message_dialog(app, _("Couldn't create a Spice session"));
+        if (!virt_viewer_app_create_session(app, "spice", &error)) {
+            g_debug("Couldn't create a Spice session");
             goto cleanup;
         }
 
@@ -1036,8 +1096,13 @@ remote_viewer_start(VirtViewerApp *app)
 #endif
 retry_dialog:
         if (priv->open_recent_dialog) {
-            if (connect_dialog(&guri) != 0)
+            if (!remote_viewer_connect_dialog(&guri)) {
+                g_set_error_literal(&error,
+                            VIRT_VIEWER_ERROR, VIRT_VIEWER_ERROR_CANCELLED,
+                            _("No connection was chosen"));
+                g_propagate_error(err, error);
                 return FALSE;
+            }
             g_object_set(app, "guri", guri, NULL);
         } else
             g_object_get(app, "guri", &guri, NULL);
@@ -1052,46 +1117,54 @@ retry_dialog:
             vvfile = virt_viewer_file_new(path, &error);
             g_free(path);
             if (error) {
-                virt_viewer_app_simple_message_dialog(app, _("Invalid file %s"), guri);
+                g_prefix_error(&error, _("Invalid file %s: "), guri);
                 g_warning("%s", error->message);
-                g_clear_error(&error);
                 goto cleanup;
             }
             g_object_get(G_OBJECT(vvfile), "type", &type, NULL);
         } else if (virt_viewer_util_extract_host(guri, &type, NULL, NULL, NULL, NULL) < 0 || type == NULL) {
-            virt_viewer_app_simple_message_dialog(app, _("Cannot determine the connection type from URI"));
+            g_set_error_literal(&error,
+                                VIRT_VIEWER_ERROR, VIRT_VIEWER_ERROR_FAILED,
+                                _("Cannot determine the connection type from URI"));
             goto cleanup;
         }
 #ifdef HAVE_OVIRT
         if (g_strcmp0(type, "ovirt") == 0) {
-            if (!create_ovirt_session(app, guri)) {
-                virt_viewer_app_simple_message_dialog(app, _("Couldn't open oVirt session"));
+            if (!create_ovirt_session(app, guri, &error)) {
+                g_prefix_error(&error, _("Couldn't open oVirt session: "));
                 goto cleanup;
             }
         } else
 #endif
         {
-            if (virt_viewer_app_create_session(app, type) < 0) {
-                virt_viewer_app_simple_message_dialog(app, _("Couldn't create a session for this type: %s"), type);
+            if (!virt_viewer_app_create_session(app, type, &error))
                 goto cleanup;
-            }
         }
 
         virt_viewer_session_set_file(virt_viewer_app_get_session(app), vvfile);
+#ifdef HAVE_OVIRT
+        if (vvfile != NULL) {
+            OvirtForeignMenu *ovirt_menu;
+            ovirt_menu = ovirt_foreign_menu_new_from_file(vvfile);
+            if (ovirt_menu != NULL) {
+                virt_viewer_app_set_ovirt_foreign_menu(app, ovirt_menu);
+            }
+        }
+#endif
 
         if (!virt_viewer_app_initial_connect(app, &error)) {
-            const gchar *msg = error ? error->message :
-                _("Failed to initiate connection");
-
-            virt_viewer_app_simple_message_dialog(app, msg);
-            g_clear_error(&error);
+            if (error == NULL) {
+                g_set_error_literal(&error,
+                                    VIRT_VIEWER_ERROR, VIRT_VIEWER_ERROR_FAILED,
+                                    _("Failed to initiate connection"));
+            }
             goto cleanup;
         }
 #ifdef HAVE_SPICE_GTK
     }
 #endif
 
-    ret = VIRT_VIEWER_APP_CLASS(remote_viewer_parent_class)->start(app);
+    ret = VIRT_VIEWER_APP_CLASS(remote_viewer_parent_class)->start(app, &error);
 
 cleanup:
     g_clear_object(&file);
@@ -1102,8 +1175,11 @@ cleanup:
     type = NULL;
 
     if (!ret && priv->open_recent_dialog) {
+        g_clear_error(&error);
         goto retry_dialog;
     }
+    if (error != NULL)
+        g_propagate_error(err, error);
 
     return ret;
 }
