@@ -102,12 +102,26 @@ gchar* virt_viewer_file_get_proxy(VirtViewerFile* self);
 void virt_viewer_file_set_proxy(VirtViewerFile* self, const gchar* value);
 gchar* virt_viewer_file_get_version(VirtViewerFile* self);
 void virt_viewer_file_set_version(VirtViewerFile* self, const gchar* value);
+GHashTable* virt_viewer_file_get_versions(VirtViewerFile* self);
+void virt_viewer_file_set_versions(VirtViewerFile* self, GHashTable *version_table);
+gchar* virt_viewer_file_get_version_url(VirtViewerFile* self);
+void virt_viewer_file_set_version_url(VirtViewerFile* self, const gchar* value);
 gchar** virt_viewer_file_get_secure_channels(VirtViewerFile* self, gsize* length);
 void virt_viewer_file_set_secure_channels(VirtViewerFile* self, const gchar* const* value, gsize length);
 gint virt_viewer_file_get_delete_this_file(VirtViewerFile* self);
 void virt_viewer_file_set_delete_this_file(VirtViewerFile* self, gint value);
 gchar* virt_viewer_file_get_secure_attention(VirtViewerFile* self);
 void virt_viewer_file_set_secure_attention(VirtViewerFile* self, const gchar* value);
+gint virt_viewer_file_get_ovirt_admin(VirtViewerFile* self);
+void virt_viewer_file_set_ovirt_admin(VirtViewerFile* self, gint value);
+gchar* virt_viewer_file_get_ovirt_host(VirtViewerFile* self);
+void virt_viewer_file_set_ovirt_host(VirtViewerFile* self, const gchar* value);
+gchar* virt_viewer_file_get_ovirt_vm_guid(VirtViewerFile* self);
+void virt_viewer_file_set_ovirt_vm_guid(VirtViewerFile* self, const gchar* value);
+gchar* virt_viewer_file_get_ovirt_jsessionid(VirtViewerFile* self);
+void virt_viewer_file_set_ovirt_jsessionid(VirtViewerFile* self, const gchar* value);
+gchar* virt_viewer_file_get_ovirt_ca(VirtViewerFile* self);
+void virt_viewer_file_set_ovirt_ca(VirtViewerFile* self, const gchar* value);
 
 G_END_DECLS
 
