@@ -44,10 +44,6 @@
 #include "gbinding.c"
 #endif
 
-#ifndef SPICE_GTK_CHECK_VERSION
-#define SPICE_GTK_CHECK_VERSION(x, y, z) 0
-#endif
-
 G_DEFINE_TYPE (VirtViewerSessionSpice, virt_viewer_session_spice, VIRT_VIEWER_TYPE_SESSION)
 
 
@@ -70,6 +66,7 @@ enum {
     PROP_0,
     PROP_SPICE_SESSION,
     PROP_SW_SMARTCARD_READER,
+    PROP_MAIN_WINDOW
 };
 
 
@@ -88,7 +85,24 @@ static void virt_viewer_session_spice_channel_destroy(SpiceSession *s,
 static void virt_viewer_session_spice_smartcard_insert(VirtViewerSession *session);
 static void virt_viewer_session_spice_smartcard_remove(VirtViewerSession *session);
 static gboolean virt_viewer_session_spice_fullscreen_auto_conf(VirtViewerSessionSpice *self);
-static void virt_viewer_session_spice_apply_monitor_geometry(VirtViewerSession *self, GdkRectangle *monitors, guint nmonitors);
+static void virt_viewer_session_spice_apply_monitor_geometry(VirtViewerSession *self, GHashTable *monitors);
+
+static void virt_viewer_session_spice_clear_displays(VirtViewerSessionSpice *self)
+{
+    SpiceSession *session = self->priv->session;
+    GList *l;
+    GList *channels;
+
+    channels = spice_session_get_channels(session);
+    for (l = channels; l != NULL; l = l->next) {
+        SpiceChannel *channel = SPICE_CHANNEL(l->data);
+
+        g_object_set_data(G_OBJECT(channel), "virt-viewer-displays", NULL);
+    }
+    g_list_free(channels);
+    virt_viewer_session_clear_displays(VIRT_VIEWER_SESSION(self));
+}
+
 
 static void
 virt_viewer_session_spice_get_property(GObject *object, guint property_id,
@@ -104,6 +118,9 @@ virt_viewer_session_spice_get_property(GObject *object, guint property_id,
     case PROP_SW_SMARTCARD_READER:
         g_value_set_boolean(value, priv->has_sw_smartcard_reader);
         break;
+    case PROP_MAIN_WINDOW:
+        g_value_set_object(value, self->priv->main_window);
+        break;
     default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
     }
@@ -113,7 +130,12 @@ static void
 virt_viewer_session_spice_set_property(GObject *object, guint property_id,
                                        const GValue *value G_GNUC_UNUSED, GParamSpec *pspec)
 {
+    VirtViewerSessionSpice *self = VIRT_VIEWER_SESSION_SPICE(object);
+
     switch (property_id) {
+    case PROP_MAIN_WINDOW:
+        self->priv->main_window = g_value_dup_object(value);
+        break;
     default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
     }
@@ -132,8 +154,7 @@ virt_viewer_session_spice_dispose(GObject *obj)
 
     spice->priv->audio = NULL;
 
-    if (spice->priv->main_window)
-        g_object_unref(spice->priv->main_window);
+    g_clear_object(&spice->priv->main_window);
 
     G_OBJECT_CLASS(virt_viewer_session_spice_parent_class)->dispose(obj);
 }
@@ -145,6 +166,75 @@ virt_viewer_session_spice_mime_type(VirtViewerSession *self G_GNUC_UNUSED)
     return "application/x-spice";
 }
 
+static gboolean
+virt_viewer_session_spice_can_share_folder(VirtViewerSession *session)
+{
+    VirtViewerSessionSpice *self = VIRT_VIEWER_SESSION_SPICE(session);
+
+    return spice_session_has_channel_type(self->priv->session, SPICE_CHANNEL_WEBDAV);
+}
+
+static gboolean
+virt_viewer_session_spice_can_retry_auth(VirtViewerSession *session G_GNUC_UNUSED)
+{
+    return TRUE;
+}
+
+static void
+create_spice_session(VirtViewerSessionSpice *self);
+
+static void
+property_notify_do_auto_conf(GObject *gobject G_GNUC_UNUSED,
+                             GParamSpec *pspec G_GNUC_UNUSED,
+                             VirtViewerSessionSpice *self)
+{
+    virt_viewer_session_spice_fullscreen_auto_conf(self);
+}
+
+static void
+update_share_folder(VirtViewerSessionSpice *self)
+{
+    gboolean share;
+    SpiceSession *session = self->priv->session;
+    GList *l, *channels;
+
+    g_object_get(self, "share-folder", &share, NULL);
+
+    channels = spice_session_get_channels(session);
+    for (l = channels; l != NULL; l = l->next) {
+        SpiceChannel *channel = l->data;
+
+        if (!SPICE_IS_WEBDAV_CHANNEL(channel))
+            continue;
+
+        if (share)
+            spice_channel_connect(channel);
+        else
+            spice_channel_disconnect(channel, SPICE_CHANNEL_NONE);
+    }
+
+    g_list_free(channels);
+}
+
+static void
+virt_viewer_session_spice_constructed(GObject *obj)
+{
+    VirtViewerSessionSpice *self = VIRT_VIEWER_SESSION_SPICE(obj);
+
+    create_spice_session(self);
+
+    virt_viewer_signal_connect_object(virt_viewer_session_get_app(VIRT_VIEWER_SESSION(self)),
+                                      "notify::fullscreen",
+                                      G_CALLBACK(property_notify_do_auto_conf),
+                                      self, 0);
+
+    virt_viewer_signal_connect_object(self, "notify::share-folder",
+                                      G_CALLBACK(update_share_folder), self,
+                                      G_CONNECT_SWAPPED);
+
+    G_OBJECT_CLASS(virt_viewer_session_spice_parent_class)->constructed(obj);
+}
+
 static void
 virt_viewer_session_spice_class_init(VirtViewerSessionSpiceClass *klass)
 {
@@ -154,6 +244,7 @@ virt_viewer_session_spice_class_init(VirtViewerSessionSpiceClass *klass)
     oclass->get_property = virt_viewer_session_spice_get_property;
     oclass->set_property = virt_viewer_session_spice_set_property;
     oclass->dispose = virt_viewer_session_spice_dispose;
+    oclass->constructed = virt_viewer_session_spice_constructed;
 
     dclass->close = virt_viewer_session_spice_close;
     dclass->open_fd = virt_viewer_session_spice_open_fd;
@@ -165,6 +256,8 @@ virt_viewer_session_spice_class_init(VirtViewerSessionSpiceClass *klass)
     dclass->smartcard_remove = virt_viewer_session_spice_smartcard_remove;
     dclass->mime_type = virt_viewer_session_spice_mime_type;
     dclass->apply_monitor_geometry = virt_viewer_session_spice_apply_monitor_geometry;
+    dclass->can_share_folder = virt_viewer_session_spice_can_share_folder;
+    dclass->can_retry_auth = virt_viewer_session_spice_can_retry_auth;
 
     g_type_class_add_private(klass, sizeof(VirtViewerSessionSpicePrivate));
 
@@ -176,6 +269,14 @@ virt_viewer_session_spice_class_init(VirtViewerSessionSpiceClass *klass)
                                                         SPICE_TYPE_SESSION,
                                                         G_PARAM_READABLE |
                                                         G_PARAM_STATIC_STRINGS));
+    g_object_class_install_property(oclass,
+                                    PROP_MAIN_WINDOW,
+                                    g_param_spec_object("main-window",
+                                                        "main window",
+                                                        "Main Window",
+                                                        GTK_TYPE_WINDOW,
+                                                        G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY |
+                                                        G_PARAM_STATIC_STRINGS));
     g_object_class_override_property(oclass,
                                      PROP_SW_SMARTCARD_READER,
                                      "software-smartcard-reader");
@@ -233,6 +334,51 @@ static void reader_removed_cb(SpiceSmartcardManager *manager G_GNUC_UNUSED,
     }
 }
 
+#define UUID_LEN 16
+static void
+uuid_changed(GObject *gobject G_GNUC_UNUSED,
+             GParamSpec *pspec G_GNUC_UNUSED,
+             VirtViewerSessionSpice *self)
+{
+    guint8* uuid = NULL;
+    VirtViewerApp* app = virt_viewer_session_get_app(VIRT_VIEWER_SESSION(self));
+
+    g_object_get(self->priv->session, "uuid", &uuid, NULL);
+    if (uuid) {
+        int i;
+        gboolean uuid_empty = TRUE;
+
+        for (i = 0; i < UUID_LEN; i++) {
+            if (uuid[i] != 0) {
+                uuid_empty = FALSE;
+                break;
+            }
+        }
+
+        if (!uuid_empty) {
+            gchar *uuid_str = spice_uuid_to_string(uuid);
+            g_object_set(app, "uuid", uuid_str, NULL);
+            g_free(uuid_str);
+        }
+    }
+
+    virt_viewer_session_spice_fullscreen_auto_conf(self);
+}
+
+static void
+name_changed(GObject *gobject G_GNUC_UNUSED,
+              GParamSpec *pspec G_GNUC_UNUSED,
+              VirtViewerSessionSpice *self)
+{
+    gchar *name = NULL;
+    VirtViewerApp *app = virt_viewer_session_get_app(VIRT_VIEWER_SESSION(self));
+
+    g_object_get(self->priv->session, "name", &name, NULL);
+
+    g_object_set(app, "guest-name", name, NULL);
+    g_free(name);
+}
+
 static void
 create_spice_session(VirtViewerSessionSpice *self)
 {
@@ -283,6 +429,21 @@ create_spice_session(VirtViewerSessionSpice *self)
         }
         g_list_free(readers);
     }
+
+    /* notify::uuid is guaranteed to be emitted during connection startup even
+     * if the server is too old to support sending uuid */
+    virt_viewer_signal_connect_object(self->priv->session, "notify::uuid",
+                                      G_CALLBACK(uuid_changed), self, 0);
+    virt_viewer_signal_connect_object(self->priv->session, "notify::name",
+                                      G_CALLBACK(name_changed), self, 0);
+
+    g_object_bind_property(self->priv->session, "shared-dir",
+                           self, "shared-folder",
+                           G_BINDING_BIDIRECTIONAL|G_BINDING_SYNC_CREATE);
+    g_object_bind_property(self->priv->session, "share-dir-ro",
+                           self, "share-folder-ro",
+                           G_BINDING_BIDIRECTIONAL|G_BINDING_SYNC_CREATE);
+
 }
 
 static void
@@ -294,7 +455,7 @@ virt_viewer_session_spice_close(VirtViewerSession *session)
 
     g_object_add_weak_pointer(G_OBJECT(self), (gpointer*)&self);
 
-    virt_viewer_session_clear_displays(session);
+    virt_viewer_session_spice_clear_displays(self);
 
     if (self->priv->session) {
         spice_session_disconnect(self->priv->session);
@@ -490,7 +651,7 @@ virt_viewer_session_spice_channel_open_fd_request(SpiceChannel *channel,
 }
 
 static void
-virt_viewer_session_spice_main_channel_event(SpiceChannel *channel G_GNUC_UNUSED,
+virt_viewer_session_spice_main_channel_event(SpiceChannel *channel,
                                              SpiceChannelEvent event,
                                              VirtViewerSession *session)
 {
@@ -509,7 +670,7 @@ virt_viewer_session_spice_main_channel_event(SpiceChannel *channel G_GNUC_UNUSED
     case SPICE_CHANNEL_CLOSED:
         g_debug("main channel: closed");
         /* Ensure the other channels get closed too */
-        virt_viewer_session_clear_displays(session);
+        virt_viewer_session_spice_clear_displays(self);
         if (self->priv->session)
             spice_session_disconnect(self->priv->session);
         break;
@@ -521,17 +682,15 @@ virt_viewer_session_spice_main_channel_event(SpiceChannel *channel G_GNUC_UNUSED
         const GError *error = NULL;
         g_debug("main channel: auth failure (wrong username/password?)");
 
-#if SPICE_GTK_CHECK_VERSION(0, 26, 0)
         {
             error = spice_channel_get_error(channel);
             username_required = g_error_matches(error,
                                                 SPICE_CLIENT_ERROR,
                                                 SPICE_CLIENT_ERROR_AUTH_NEEDS_PASSWORD_AND_USERNAME);
         }
-#endif
 
         if (self->priv->pass_try > 0)
-            g_signal_emit_by_name(session, "session-auth-failed",
+            g_signal_emit_by_name(session, "session-auth-refused",
                                   error != NULL ? error->message : _("Invalid password"));
         self->priv->pass_try++;
 
@@ -569,7 +728,6 @@ virt_viewer_session_spice_main_channel_event(SpiceChannel *channel G_GNUC_UNUSED
         break;
     }
     case SPICE_CHANNEL_ERROR_CONNECT:
-#if SPICE_GTK_CHECK_VERSION(0, 23, 21)
     {
         const GError *error = spice_channel_get_error(channel);
 
@@ -591,18 +749,14 @@ virt_viewer_session_spice_main_channel_event(SpiceChannel *channel G_GNUC_UNUSED
                 spice_session_connect(self->priv->session);
             }
         } else {
-            g_signal_emit_by_name(session, "session-disconnected", error ? error->message : NULL);
+            virt_viewer_session_spice_channel_destroy(NULL, channel, session);
         }
-    }
-#else
-        g_debug("main channel: failed to connect");
-        g_signal_emit_by_name(session, "session-disconnected", NULL);
-#endif
         break;
+    }
     case SPICE_CHANNEL_ERROR_IO:
     case SPICE_CHANNEL_ERROR_LINK:
     case SPICE_CHANNEL_ERROR_TLS:
-        g_signal_emit_by_name(session, "session-disconnected", NULL);
+        virt_viewer_session_spice_channel_destroy(NULL, channel, session);
         break;
     default:
         g_warning("unhandled spice main channel event: %d", event);
@@ -660,12 +814,17 @@ agent_connected_changed(SpiceChannel *cmain G_GNUC_UNUSED,
                         GParamSpec *pspec G_GNUC_UNUSED,
                         VirtViewerSessionSpice *self)
 {
+    gboolean agent_connected;
+
     // this will force refresh of application menu
     g_signal_emit_by_name(self, "session-display-updated");
 
-    /* this will force update displays geometry when the agent has connected
-     * after the application (eg: rebooting the guest) */
-    virt_viewer_session_update_displays_geometry(VIRT_VIEWER_SESSION(self));
+    g_object_get(cmain, "agent-connected", &agent_connected, NULL);
+    if (agent_connected) {
+        /* this will force update displays geometry when the agent has connected
+         * after the application (eg: rebooting the guest) */
+        virt_viewer_session_update_displays_geometry(VIRT_VIEWER_SESSION(self));
+    }
 }
 
 static void
@@ -679,6 +838,16 @@ destroy_display(gpointer data)
     g_object_unref(display);
 }
 
+static gboolean
+display_is_in_fullscreen_mode(VirtViewerSessionSpice *self,
+                              VirtViewerDisplay *display)
+{
+    gint nth = virt_viewer_display_get_nth(display);
+    VirtViewerApp *app = virt_viewer_session_get_app(VIRT_VIEWER_SESSION(self));
+
+    return virt_viewer_app_get_initial_monitor_for_display(app, nth) != -1;
+}
+
 static void
 virt_viewer_session_spice_display_monitors(SpiceChannel *channel,
                                            GParamSpec *pspec G_GNUC_UNUSED,
@@ -688,6 +857,8 @@ virt_viewer_session_spice_display_monitors(SpiceChannel *channel,
     GPtrArray *displays = NULL;
     GtkWidget *display;
     guint i, monitors_max;
+    gboolean fullscreen_mode =
+        virt_viewer_app_get_fullscreen(virt_viewer_session_get_app(VIRT_VIEWER_SESSION(self)));
 
     g_object_get(channel,
                  "monitors", &monitors,
@@ -710,25 +881,38 @@ virt_viewer_session_spice_display_monitors(SpiceChannel *channel,
         display = g_ptr_array_index(displays, i);
         if (display == NULL) {
             display = virt_viewer_display_spice_new(self, channel, i);
-            g_debug("creating spice display (#:%d)", i);
+            g_debug("creating spice display (#:%d)",
+                    virt_viewer_display_get_nth(VIRT_VIEWER_DISPLAY(display)));
             g_ptr_array_index(displays, i) = g_object_ref_sink(display);
+            virt_viewer_session_add_display(VIRT_VIEWER_SESSION(self),
+                                            VIRT_VIEWER_DISPLAY(display));
         }
-
-        virt_viewer_session_add_display(VIRT_VIEWER_SESSION(self),
-                                        VIRT_VIEWER_DISPLAY(display));
     }
 
     for (i = 0; i < monitors->len; i++) {
         SpiceDisplayMonitorConfig *monitor = &g_array_index(monitors, SpiceDisplayMonitorConfig, i);
+        gboolean disabled = monitor->width == 0 || monitor->height == 0;
         display = g_ptr_array_index(displays, monitor->id);
         g_return_if_fail(display != NULL);
 
-        if (monitor->width == 0 || monitor->height == 0)
+        if (!disabled && fullscreen_mode && self->priv->did_auto_conf &&
+            !display_is_in_fullscreen_mode(self, VIRT_VIEWER_DISPLAY(display))) {
+            g_warning("display %d should not be enabled, disabling",
+                      virt_viewer_display_get_nth(VIRT_VIEWER_DISPLAY(display)));
+            spice_main_set_display_enabled(virt_viewer_session_spice_get_main_channel(self),
+                                           virt_viewer_display_get_nth(VIRT_VIEWER_DISPLAY(display)),
+                                           FALSE);
+            disabled = TRUE;
+        }
+
+        virt_viewer_display_set_enabled(VIRT_VIEWER_DISPLAY(display), !disabled);
+
+        if (disabled)
             continue;
 
-        virt_viewer_display_set_enabled(VIRT_VIEWER_DISPLAY(display), TRUE);
-        virt_viewer_display_set_desktop_size(VIRT_VIEWER_DISPLAY(display),
-                                             monitor->width, monitor->height);
+        virt_viewer_display_spice_set_desktop(VIRT_VIEWER_DISPLAY(display),
+                                              monitor->x, monitor->y,
+                                              monitor->width, monitor->height);
     }
 
     g_clear_pointer(&monitors, g_array_unref);
@@ -798,24 +982,18 @@ virt_viewer_session_spice_channel_new(SpiceSession *s,
     self->priv->channel_count++;
 }
 
-static void
-property_notify_do_auto_conf(GObject *gobject G_GNUC_UNUSED,
-                             GParamSpec *pspec G_GNUC_UNUSED,
-                             VirtViewerSessionSpice *self)
-{
-    virt_viewer_session_spice_fullscreen_auto_conf(self);
-}
-
 static gboolean
 virt_viewer_session_spice_fullscreen_auto_conf(VirtViewerSessionSpice *self)
 {
     GdkScreen *screen = gdk_screen_get_default();
     SpiceMainChannel* cmain = virt_viewer_session_spice_get_main_channel(self);
     VirtViewerApp *app = NULL;
-    GdkRectangle *displays;
+    GHashTable *displays;
+    GHashTableIter iter;
+    gpointer key, value;
     gboolean agent_connected;
-    gint i;
-    gsize ndisplays = 0;
+    GList *initial_displays, *l;
+    guint ndisplays;
 
     /* only do auto-conf once at startup. Avoid repeating auto-conf later due to
      * agent disconnection/re-connection, etc */
@@ -845,30 +1023,35 @@ virt_viewer_session_spice_fullscreen_auto_conf(VirtViewerSessionSpice *self)
 
     spice_main_set_display_enabled(cmain, -1, FALSE);
 
-    ndisplays = virt_viewer_app_get_n_initial_displays(app);
-    g_debug("Performing full screen auto-conf, %" G_GSIZE_FORMAT " host monitors", ndisplays);
-    displays = g_new0(GdkRectangle, ndisplays);
+    initial_displays = virt_viewer_app_get_initial_displays(app);
+    ndisplays = g_list_length(initial_displays);
+    g_debug("Performing full screen auto-conf, %u host monitors", ndisplays);
+    displays = g_hash_table_new_full(g_direct_hash, g_direct_equal, NULL, g_free);
 
-    for (i = 0; i < ndisplays; i++) {
-        GdkRectangle* rect = &displays[i];
-        gint j = virt_viewer_app_get_initial_monitor_for_display(app, i);
+    for (l = initial_displays; l != NULL; l = l->next) {
+        GdkRectangle* rect = g_new0(GdkRectangle, 1);;
+        gint j = virt_viewer_app_get_initial_monitor_for_display(app, GPOINTER_TO_INT(l->data));
         if (j == -1)
             continue;
 
         gdk_screen_get_monitor_geometry(screen, j, rect);
+        g_hash_table_insert(displays, l->data, rect);
     }
 
-    virt_viewer_shift_monitors_to_origin(displays, ndisplays);
+    virt_viewer_shift_monitors_to_origin(displays);
 
-    for (i = 0; i < ndisplays; i++) {
-        GdkRectangle *rect = &displays[i];
+    g_hash_table_iter_init(&iter, displays);
+    while (g_hash_table_iter_next(&iter, &key, &value)) {
+        GdkRectangle *rect = value;
+        gint j = GPOINTER_TO_INT(key);
 
-        spice_main_set_display(cmain, i, rect->x, rect->y, rect->width, rect->height);
-        spice_main_set_display_enabled(cmain, i, TRUE);
+        spice_main_set_display(cmain, j, rect->x, rect->y, rect->width, rect->height);
+        spice_main_set_display_enabled(cmain, j, TRUE);
         g_debug("Set SPICE display %d to (%d,%d)-(%dx%d)",
-                  i, rect->x, rect->y, rect->width, rect->height);
+                  j, rect->x, rect->y, rect->width, rect->height);
     }
-    g_free(displays);
+    g_list_free(initial_displays);
+    g_hash_table_unref(displays);
 
     spice_main_send_monitor_config(cmain);
     self->priv->did_auto_conf = TRUE;
@@ -882,12 +1065,15 @@ virt_viewer_session_spice_channel_destroy(G_GNUC_UNUSED SpiceSession *s,
 {
     VirtViewerSessionSpice *self = VIRT_VIEWER_SESSION_SPICE(session);
     int id;
+    const GError *error;
 
     g_return_if_fail(self != NULL);
 
     g_object_get(channel, "channel-id", &id, NULL);
     g_debug("Destroy SPICE channel %s %d", g_type_name(G_OBJECT_TYPE(channel)), id);
 
+    error = spice_channel_get_error(channel);
+
     if (SPICE_IS_MAIN_CHANNEL(channel)) {
         g_debug("zap main channel");
         if (channel == SPICE_CHANNEL(self->priv->main_channel))
@@ -913,52 +1099,7 @@ virt_viewer_session_spice_channel_destroy(G_GNUC_UNUSED SpiceSession *s,
 
     self->priv->channel_count--;
     if (self->priv->channel_count == 0)
-        g_signal_emit_by_name(self, "session-disconnected", NULL);
-}
-
-#define UUID_LEN 16
-static void
-uuid_changed(GObject *gobject G_GNUC_UNUSED,
-             GParamSpec *pspec G_GNUC_UNUSED,
-             VirtViewerSessionSpice *self)
-{
-    guint8* uuid = NULL;
-    VirtViewerApp* app = virt_viewer_session_get_app(VIRT_VIEWER_SESSION(self));
-
-    g_object_get(self->priv->session, "uuid", &uuid, NULL);
-    if (uuid) {
-        int i;
-        gboolean uuid_empty = TRUE;
-
-        for (i = 0; i < UUID_LEN; i++) {
-            if (uuid[i] != 0) {
-                uuid_empty = FALSE;
-                break;
-            }
-        }
-
-        if (!uuid_empty) {
-            gchar *uuid_str = spice_uuid_to_string(uuid);
-            g_object_set(app, "uuid", uuid_str, NULL);
-            g_free(uuid_str);
-        }
-    }
-
-    virt_viewer_session_spice_fullscreen_auto_conf(self);
-}
-
-static void
-name_changed(GObject *gobject G_GNUC_UNUSED,
-              GParamSpec *pspec G_GNUC_UNUSED,
-              VirtViewerSessionSpice *self)
-{
-    gchar *name = NULL;
-    VirtViewerApp *app = virt_viewer_session_get_app(VIRT_VIEWER_SESSION(self));
-
-    g_object_get(self->priv->session, "name", &name, NULL);
-
-    g_object_set(app, "guest-name", name, NULL);
-    g_free(name);
+        g_signal_emit_by_name(self, "session-disconnected", error ? error->message : NULL);
 }
 
 VirtViewerSession *
@@ -966,21 +1107,10 @@ virt_viewer_session_spice_new(VirtViewerApp *app, GtkWindow *main_window)
 {
     VirtViewerSessionSpice *self;
 
-    self = g_object_new(VIRT_VIEWER_TYPE_SESSION_SPICE, "app", app, NULL);
-
-    create_spice_session(self);
-    self->priv->main_window = g_object_ref(main_window);
-
-    virt_viewer_signal_connect_object(app, "notify::fullscreen",
-                                      G_CALLBACK(property_notify_do_auto_conf), self, 0);
-
-    /* notify::uuid is guaranteed to be emitted during connection startup even
-     * if the server is too old to support sending uuid */
-    virt_viewer_signal_connect_object(self->priv->session, "notify::uuid",
-                                      G_CALLBACK(uuid_changed), self, 0);
-    virt_viewer_signal_connect_object(self->priv->session, "notify::name",
-                                      G_CALLBACK(name_changed), self, 0);
-
+    self = g_object_new(VIRT_VIEWER_TYPE_SESSION_SPICE,
+                        "app", app,
+                        "main-window", main_window,
+                        NULL);
     return VIRT_VIEWER_SESSION(self);
 }
 
@@ -1004,14 +1134,17 @@ virt_viewer_session_spice_smartcard_remove(VirtViewerSession *session G_GNUC_UNU
     spice_smartcard_manager_remove_card(spice_smartcard_manager_get());
 }
 
-void
-virt_viewer_session_spice_apply_monitor_geometry(VirtViewerSession *session, GdkRectangle *monitors, guint nmonitors)
+static void
+virt_viewer_session_spice_apply_monitor_geometry(VirtViewerSession *session, GHashTable *monitors)
 {
-    guint i;
+    GHashTableIter iter;
+    gpointer key = NULL, value = NULL;
     VirtViewerSessionSpice *self = VIRT_VIEWER_SESSION_SPICE(session);
 
-    for (i = 0; i < nmonitors; i++) {
-        GdkRectangle* rect = &monitors[i];
+    g_hash_table_iter_init(&iter, monitors);
+    while (g_hash_table_iter_next(&iter, &key, &value)) {
+        gint i = GPOINTER_TO_INT(key);
+        GdkRectangle* rect = value;
 
         spice_main_set_display(self->priv->main_channel, i, rect->x,
                                rect->y, rect->width, rect->height);
