@@ -1,7 +1,7 @@
 /*
  * Virt Viewer: A virtual machine console viewer
  *
- * Copyright (C) 2007-2012 Red Hat, Inc.
+ * Copyright (C) 2007-2015 Red Hat, Inc.
  * Copyright (C) 2009-2012 Daniel P. Berrange
  *
  * This program is free software; you can redistribute it and/or modify
@@ -117,7 +117,7 @@ virt_viewer_util_extract_host(const char *uristr,
     g_return_val_if_fail(uri != NULL, 1);
 
     if (host) {
-        if (!uri || !uri->server) {
+        if (!uri->server) {
             *host = g_strdup("localhost");
         } else {
             if (uri->server[0] == '[') {
@@ -323,7 +323,7 @@ void virt_viewer_util_init(const char *appname)
 
     g_set_application_name(appname);
 
-    g_log_set_default_handler(log_handler, NULL);
+    g_log_set_handler(G_LOG_DOMAIN, G_LOG_LEVEL_MASK, log_handler, NULL);
 }
 
 static gchar *
@@ -439,25 +439,24 @@ spice_hotkey_to_gtk_accelerator(const gchar *key)
     return accel;
 }
 
-/**
- * virt_viewer_compare_version:
- * @s1: a version-like string
- * @s2: a version-like string
- *
- * Compare two version-like strings: 1.1 > 1.0, 1.0.1 > 1.0, 1.10 > 1.7...
- *
- * String suffix (1.0rc1 etc) are not accepted, and will return 0.
- *
- * Returns: negative value if s1 < s2; zero if s1 = s2; positive value if s1 > s2.
- **/
-gint
+static gboolean str_is_empty(const gchar *str)
+{
+  return ((str == NULL) || (str[0] == '\0'));
+}
+
+static gint
 virt_viewer_compare_version(const gchar *s1, const gchar *s2)
 {
     gint i, retval = 0;
     gchar **v1, **v2;
 
-    g_return_val_if_fail(s1 != NULL, 0);
-    g_return_val_if_fail(s2 != NULL, 0);
+    if (str_is_empty(s1) && str_is_empty(s2)) {
+      return 0;
+    } else if (str_is_empty(s1)) {
+      return -1;
+    } else if (str_is_empty(s2)) {
+      return 1;
+    }
 
     v1 = g_strsplit(s1, ".", -1);
     v2 = g_strsplit(s2, ".", -1);
@@ -473,7 +472,7 @@ virt_viewer_compare_version(const gchar *s1, const gchar *s2)
 
         g_return_val_if_fail(e1 && e2, 0);
         if (*e1 || *e2) {
-            g_warning("the version string contains suffix");
+            g_warning("the version string contains a suffix");
             goto end;
         }
     }
@@ -488,6 +487,163 @@ end:
     g_strfreev(v2);
     return retval;
 }
+
+/**
+ * virt_viewer_compare_buildid:
+ * @s1: a version-like string
+ * @s2: a version-like string
+ *
+ * Compare two buildid strings: 1.1-1 > 1.0-1, 1.0-2 > 1.0-1, 1.10 > 1.7...
+ *
+ * String suffix (1.0rc1 etc) are not accepted, and will return 0.
+ *
+ * Returns: negative value if s1 < s2; zero if s1 = s2; positive value if s1 > s2.
+ **/
+gint
+virt_viewer_compare_buildid(const gchar *s1, const gchar *s2)
+{
+    int ret = 0;
+    GStrv split1 = NULL;
+    GStrv split2 = NULL;
+
+    split1 = g_strsplit(s1, "-", 2);
+    split2 = g_strsplit(s2, "-", 2);
+    if ((split1 == NULL) || (split2 == NULL)) {
+      goto end;
+    }
+    /* Compare versions */
+    ret = virt_viewer_compare_version(split1[0], split2[0]);
+    if (ret != 0) {
+        goto end;
+    }
+    if ((split1[0] == NULL) || (split2[0] == NULL)) {
+      goto end;
+    }
+
+    /* Compare -release */
+    ret = virt_viewer_compare_version(split1[1], split2[1]);
+
+end:
+    g_strfreev(split1);
+    g_strfreev(split2);
+
+    return ret;
+}
+
+/* simple sorting of monitors. Primary sort left-to-right, secondary sort from
+ * top-to-bottom, finally by monitor id */
+static int
+displays_cmp(const void *p1, const void *p2, gpointer user_data)
+{
+    guint diff;
+    GHashTable *displays = user_data;
+    guint i = *(guint*)p1;
+    guint j = *(guint*)p2;
+    GdkRectangle *m1 = g_hash_table_lookup(displays, GINT_TO_POINTER(i));
+    GdkRectangle *m2 = g_hash_table_lookup(displays, GINT_TO_POINTER(j));
+    diff = m1->x - m2->x;
+    if (diff == 0)
+        diff = m1->y - m2->y;
+    if (diff == 0)
+        diff = i - j;
+
+    return diff;
+}
+
+static void find_max_id(gpointer key,
+                        gpointer value G_GNUC_UNUSED,
+                        gpointer user_data)
+{
+    guint *max_id = user_data;
+    guint id = GPOINTER_TO_INT(key);
+    *max_id = MAX(*max_id, id);
+}
+
+void
+virt_viewer_align_monitors_linear(GHashTable *displays)
+{
+    gint i, x = 0;
+    guint *sorted_displays;
+    guint max_id = 0;
+    guint ndisplays = 0;
+    GHashTableIter iter;
+    gpointer key, value;
+
+    g_return_if_fail(displays != NULL);
+
+    if (g_hash_table_size(displays) == 0)
+        return;
+
+    g_hash_table_foreach(displays, find_max_id, &max_id);
+    ndisplays = max_id + 1;
+
+    sorted_displays = g_new0(guint, ndisplays);
+
+    g_hash_table_iter_init(&iter, displays);
+    while (g_hash_table_iter_next(&iter, &key, &value))
+        sorted_displays[GPOINTER_TO_INT(key)] = GPOINTER_TO_INT(key);
+
+    g_qsort_with_data(sorted_displays, ndisplays, sizeof(guint), displays_cmp, displays);
+
+    /* adjust monitor positions so that there's no gaps or overlap between
+     * monitors */
+    for (i = 0; i < ndisplays; i++) {
+        guint nth = sorted_displays[i];
+        g_assert(nth < ndisplays);
+        GdkRectangle *rect = g_hash_table_lookup(displays, GINT_TO_POINTER(nth));
+        rect->x = x;
+        rect->y = 0;
+        x += rect->width;
+    }
+    g_free(sorted_displays);
+}
+
+/* Shift all displays so that the monitor origin is at (0,0). This reduces the
+ * size of the screen that will be required on the guest when all client
+ * monitors are fullscreen but do not begin at the origin. For example, instead
+ * of sending down the following configuration:
+ *   1280x1024+4240+0
+ * (which implies that the guest screen must be at least 5520x1024), we'd send
+ *   1280x1024+0+0
+ * (which implies the guest screen only needs to be 1280x1024). The first
+ * version might fail if the guest video memory is not large enough to handle a
+ * screen of that size.
+ */
+void
+virt_viewer_shift_monitors_to_origin(GHashTable *displays)
+{
+    gint xmin = G_MAXINT;
+    gint ymin = G_MAXINT;
+    GHashTableIter iter;
+    gpointer key, value;
+
+    if (g_hash_table_size(displays) == 0)
+        return;
+
+    g_hash_table_iter_init(&iter, displays);
+    while (g_hash_table_iter_next(&iter, &key, &value)) {
+        GdkRectangle *display = value;
+        if (display->width > 0 && display->height > 0) {
+            xmin = MIN(xmin, display->x);
+            ymin = MIN(ymin, display->y);
+        }
+    }
+    g_return_if_fail(xmin < G_MAXINT && ymin < G_MAXINT);
+
+    if (xmin > 0 || ymin > 0) {
+        g_debug("%s: Shifting all monitors by (%i, %i)", G_STRFUNC, xmin, ymin);
+        g_hash_table_iter_init(&iter, displays);
+        while (g_hash_table_iter_next(&iter, &key, &value)) {
+            GdkRectangle *display = value;
+            if (display->width > 0 && display->height > 0) {
+                display->x -= xmin;
+                display->y -= ymin;
+            }
+        }
+    }
+}
+
+
 /*
  * Local variables:
  *  c-indent-level: 4
