@@ -32,6 +32,8 @@ G_BEGIN_DECLS
 
 #define MIN_ZOOM_LEVEL 10
 #define MAX_ZOOM_LEVEL 400
+#define NORMAL_ZOOM_LEVEL 100
+#define ZOOM_STEP 10
 
 #define VIRT_VIEWER_TYPE_WINDOW virt_viewer_window_get_type()
 
@@ -67,7 +69,9 @@ GtkWindow* virt_viewer_window_get_window (VirtViewerWindow* window);
 VirtViewerNotebook* virt_viewer_window_get_notebook (VirtViewerWindow* window);
 void virt_viewer_window_set_display(VirtViewerWindow *self, VirtViewerDisplay *display);
 VirtViewerDisplay* virt_viewer_window_get_display(VirtViewerWindow *self);
+void virt_viewer_window_set_menu_displays_sensitive(VirtViewerWindow *self, gboolean sensitive);
 void virt_viewer_window_set_usb_options_sensitive(VirtViewerWindow *self, gboolean sensitive);
+void virt_viewer_window_set_menus_sensitive(VirtViewerWindow *self, gboolean sensitive);
 void virt_viewer_window_update_title(VirtViewerWindow *self);
 void virt_viewer_window_show(VirtViewerWindow *self);
 void virt_viewer_window_hide(VirtViewerWindow *self);
